import polars as pl
import polars._typing as pt


def into_expr(w: pt.IntoExprColumn) -> pl.Expr:
    """
    Converts a string (column name) or Polars series into an expression.
    """
    if isinstance(w, str):
        return pl.col(w)

    elif isinstance(w, pl.Series):
        return pl.lit(w)

    elif isinstance(w, pl.Expr):
        return w

    else:
        raise ValueError


def normalize(x: pl.Expr) -> pl.Expr:
    """
    Normalizes an expression so that it sums to one.
    """
    return x / x.sum()


def match_name(x: pl.Expr, col_to_match: pl.Expr, *, fallback="literal") -> pl.Expr:
    """
    Changes the name of an expression to match the name of another expression.
    """
    return x.alias(
        col_to_match.meta.output_name(raise_if_undetermined=False) or fallback
    )



from typing import Literal

import polars as pl

from polars_utils import match_name
from polars_utils.weights import Weight, into_normalized_weight


def quantile(
    x: pl.Expr,
    *,
    w: Weight = None,
    endpoint: Literal["left", "right", "midpoint"] = "midpoint",
) -> pl.Expr:
    """
    Computes (weighted) quantiles of a column.
    """
    right_quantiles = into_normalized_weight(w).sort_by(x).cum_sum()  # ends with a 1
    left_quantiles = right_quantiles.shift(n=+1, fill_value=0)  # starts with a zero

    match endpoint:
        case "midpoint":
            quantiles = [right_quantiles, left_quantiles]
        case "right":
            quantiles = [right_quantiles]
        case "left":
            quantiles = [left_quantiles]
        case _:
            raise ValueError

    return (
        pl.mean_horizontal(*quantiles)
        .gather(x.rank(method="ordinal") - 1)
        # set name of output
        .pipe(match_name, x)
    )


def xtile(x: pl.Expr, *, n: int, w: Weight = None, label="{i}"):
    """
    Splits data into bins of roughly equal weight a la `xtile` in stata.
    """
    quantiles = x.pipe(quantile, w=w, endpoint="midpoint")

    return quantiles.cut(
        breaks=pl.linear_space(0, 1, n - 1, closed="none", eager=True).to_list(),
        labels=[label.format(i=i + 1, n=n) for i in range(n)],
    )



from typing import Optional
import polars as pl
from polars import _typing as pt


from polars_utils import into_expr
from polars_utils.stats import mean
from polars_utils.weights import Weight


def reliabiliy(
    estimates: pl.Expr,
    *,
    variances: pt.IntoExprColumn,
    w: Optional[Weight] = None,
    mu: Optional[pl.Expr] = None,
):
    # point around which we compute variance
    mu = estimates.pipe(mean, w=w) if mu is None else mu

    # calculate reliability
    signal_variance = ((estimates - mu).pow(2) - variances).pipe(mean, w=w)
    return signal_variance / into_expr(variances).add(signal_variance)


def shrink(
    estimates: pl.Expr,
    *,
    variances: pt.IntoExprColumn,
    w: Optional[Weight] = None,
    mu: Optional[pl.Expr] = None,
) -> pl.Expr:
    """
    https://libgen.li/adsa82270d055f6ee991539ac0533036e0dO9ZBZS4G
    """

    # point we shrink to
    mu = estimates.pipe(mean, w=w) if mu is None else mu

    # reliability
    rho = reliabiliy(estimates, variances=variances, w=w, mu=mu)

    # shrink towards mean based on reliability
    return estimates.mul(rho) + mu.mul(1 - rho)



import polars as pl
import numpy as np

from polars_utils import stats

x, y, w, n, c = (pl.col(c) for c in ("x", "y", "w", "n", "c"))
x_null, y_null, w_null = (pl.col(c + "_null") for c in ("x", "y", "w"))

RNG = np.random.default_rng(10498091)


def replace_with_nulls(col: pl.Expr, null_proportion=0.05):
    return (
        pl.when(
            pl.int_range(pl.len())
            .shuffle(RNG.integers(10_000))
            .lt(pl.len() * null_proportion)
        )
        .then(None)
        .otherwise(col)
    )


def create_test_data(n=10_000) -> pl.DataFrame:
    x = RNG.normal(size=n)

    data = dict(
        x=x,
        y=2 * x + RNG.normal(size=n),
        w=RNG.uniform(0, 1, size=n),
        n=RNG.integers(0, 1000, size=n),
        c=[100] * n,
    )

    return pl.DataFrame(data).with_columns(
        pl.col("x", "y", "w").pipe(replace_with_nulls).name.suffix("_null"),
    )


def test_mean():
    df = create_test_data()

    assert np.allclose(
        df.select(x.pipe(stats.mean))[0, 0],
        np.average(df["x"]),
    ), "Unweighted mean differs from Numpy"

    assert np.allclose(
        df.select(x.pipe(stats.mean, w=w))[0, 0],
        np.average(df["x"], weights=df["w"]),
    ), "Weighted mean differs from Numpy"

    assert np.allclose(
        df.select(x.pipe(stats.mean, w="c"))[0, 0],
        df.select(x.pipe(stats.mean))[0, 0],
    ), "Constant weights should equal no weights"


def test_var():
    df = create_test_data()

    assert np.allclose(
        df.select(x.pipe(stats.var))[0, 0],
        df["x"].var(ddof=0),  # type: ignore
    ), "Unweighted variance differs from Numpy"

    assert np.allclose(
        df.select(x.pipe(stats.var, w=w))[0, 0],
        np.cov(df["x"], aweights=df["w"], ddof=0),
    ), "Weighted variance differs from Numpy"


def test_cov():
    df = create_test_data()

    assert np.allclose(
        df.select(x.pipe(stats.var, w=w))[0, 0],
        df.select(x.pipe(stats.cov, x, w=w))[0, 0],
    ), "Variance does not equal covariance with self"

    assert np.allclose(
        df.select(x.pipe(stats.cov, y))[0, 0],
        np.cov(df["x"], df["y"], ddof=0)[1, 0],
    ), "Unweighted covariance differs from Numpy"

    assert np.allclose(
        df.select(x.pipe(stats.cov, y, w=w))[0, 0],
        np.cov(df["x"], df["y"], aweights=df["w"], ddof=0)[1, 0],
    ), "Weighted variance differs from Numpy"


def test_cor():
    df = create_test_data()

    assert np.allclose(df.select(x.pipe(stats.cor, x))[0, 0], 1.0), (
        "Self correlation is not 1"
    )

    assert np.allclose(
        df.select(x.pipe(stats.cor, y))[0, 0],
        np.corrcoef(df["x"], df["y"])[1, 0],
    ), "Correlation differs from Numpy"


def test_nulls():
    df = create_test_data()

    assert np.allclose(
        df.select(x_null.pipe(stats.mean))[0, 0],
        df.drop_nulls().select(x.pipe(stats.mean))[0, 0],
    ), "Nulls not correctly masked in unweighted mean"

    assert np.allclose(
        df.select(x_null.pipe(stats.mean, w=w))[0, 0],
        df.drop_nulls().select(x.pipe(stats.mean, w=w))[0, 0],
    ), "Nulls not correctly masked in weighted mean"

    assert np.allclose(
        df.select(x_null.pipe(stats.var, w=w))[0, 0],
        df.drop_nulls().select(x.pipe(stats.var, w=w))[0, 0],
    ), "Nulls not correctly masked in weighted variance"

    assert np.allclose(
        df.select(x_null.pipe(stats.cov, y_null, w=w))[0, 0],
        df.drop_nulls().select(x.pipe(stats.cov, y, w=w))[0, 0],
    ), "Nulls not correctly masked in weighted covariance"



import polars as pl
from polars import _typing as pt
from typing import Iterable, Optional

from polars_utils.stats import mean
from polars_utils.weights import Weight


def absorb(
    x: pl.Expr,
    fixed_effects: Iterable[pt.IntoExpr],
    *,
    w: Weight = None,
    by: Optional[Iterable[pt.IntoExpr]] = None,
    add_back_mean=True,
):
    """
    Absorbs (categorical) fixed effects by demeaning.
    """
    # if by isn't passed, do everything together
    by = by or [pl.lit(1)]

    return (
        x
        # subtract mean
        - x.pipe(mean, w=w).over(*by, *fixed_effects)
        # add back mean within cell
        + (x.pipe(mean, w=w).over(*by) if add_back_mean else 0)
    )



from typing import Optional
import polars as pl
import polars._typing as pt

from polars_utils import into_expr
from polars_utils.weights import Weight


def mean(x: pl.Expr, *, w: Optional[Weight] = None) -> pl.Expr:
    """
    Computes the (weighted) mean of an expression.
    """
    if w is None:
        return x.mean()

    # handle null values in w (they get no weight)
    w = into_expr(w).fill_null(value=0)

    # handle null values in x by omitting those weights from the total
    sum_w = w.filter(x.is_not_null()).sum()

    return x.dot(w) / sum_w


def demean(x: pl.Expr, *, w: Optional[Weight] = None) -> pl.Expr:
    """
    Subtracts off the (weighted) mean of an expression.
    """
    return x - x.pipe(mean, w=w)


def cov(x: pl.Expr, y: pt.IntoExprColumn, *, w: Optional[Weight] = None) -> pl.Expr:
    """
    Computes the (weighted) covaraince of an expression with another expression.
    """

    return (
        (x.pipe(demean, w=w) * into_expr(y).pipe(demean, w=w))
        .pipe(mean, w=w)
        .alias("cov")
    )


def var(
    x: pl.Expr,
    *,
    w: Optional[Weight] = None,
    center_around: Optional[pl.Expr] = None,
):
    """
    Computes the (weighted) variance of an expression.
    """

    # TODO: handle bias correction:
    # https://en.wikipedia.org/wiki/Weighted_arithmetic_mean#Weighted_sample_variance
    # https://numpy.org/doc/stable/reference/generated/numpy.cov.html

    center_around = center_around or x.pipe(mean, w=w)

    return (x - center_around).pow(2).pipe(mean, w=w)


def cor(x: pl.Expr, y: pt.IntoExprColumn, *, w: Optional[Weight] = None) -> pl.Expr:
    """
    Computes the (optionally weighted) Pearson correlation coefficient.

    See: https://en.wikipedia.org/wiki/Pearson_correlation_coefficient#Weighted_correlation_coefficient
    """
    numerator = x.pipe(cov, y, w=w)
    denominator = (x.pipe(var, w=w) * into_expr(y).pipe(var, w=w)).sqrt()

    return (numerator / denominator).alias("cor")



import polars as pl
import polars._typing as pt

from polars_utils import into_expr, normalize

Weight = pt.IntoExprColumn


def into_normalized_weight(w: Weight) -> pl.Expr:
    w = into_expr(w)

    if w.meta.is_literal(allow_aliasing=True):
        raise ValueError("Literal weights not allowed")

    return w.pipe(normalize)
